from __future__ import annotations


import contextlib
import datetime
import itertools
import os.path
from pathlib import Path
import shutil
import textwrap
import typing as T
from typing import TYPE_CHECKING

from docutils import nodes
from docutils.parsers.rst import directives
from lxml import etree
from sphinx import addnodes
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain
from sphinx.util.docutils import SphinxDirective, ReferenceRole, SphinxRole
from pnb.mcl.metamodel import standard as metamodel
from pnb.mcl.io.xml import read_xmls, XmlExporter

from .toc_tree import find_model_names_in_toc_trees, SpeciPyTocTree
from dexpi.specificator.utils import uml_diagrams 

if TYPE_CHECKING:
    from pnb.mcl.metamodel import standard as metamodel
    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment
    from sphinx.util.typing import ExtensionMetadata
    
THIS_DIR = Path(__file__).parent
    
NBSP = '\u00a0'
    
SP_CSS_PATH = THIS_DIR / 'sp.css'
    
# TODO move
def multiplicity_string(lower, upper):
    
    if upper is None:
        if lower == 0:
            return '*'
        else:
            return f'{lower}..*'
    elif lower == upper:
        return str(lower)
    else:
        return f'{lower}..{upper}'



def create_automodel_documents(app: Sphinx) -> None:
    specipy_domain = app.builder.env.domains.get('sp')
    assert isinstance(specipy_domain, SpeciPyDomain)
    specipy_domain.create_automodel_documents()
    

class ElementRole(ReferenceRole):
    
    def run(self):
        
        target = self.target
        title = self.title
        explicit_title = title is not target
        
        if target.startswith('~!'):
            target = '!~' + target[2:]
            
        if target.startswith('!'):
            target = target[1:]
            link = False
        else:
            link = True

        sp = self.env.get_domain('sp')
        
        if target.startswith('~'):
            if explicit_title: # TODO == vs is
                TODO
            element = sp.get_element(target[1:])
            if not element:
                # TODO error msg
                return [nodes.inline(classes=['sp-element sp-error'], text=target[1:])], []
            title = element.name
        else:
            element = sp.get_element(target)
            if not element:
                # TODO error msg
                return [nodes.inline(classes=['sp-element sp-error'], text=target)], []
                
        if link:
            return sp.get_element_reference(
                element, title, ref_docpath=self.inliner.document.current_source)
        else:
            return [nodes.inline(classes=['sp-element'], text=title)], []

        
class NameRole(SphinxRole):
    
    def run(self):
        return [nodes.inline(classes=['sp-element'], text=self.text)], []


        
class ColumnSpec:
    
    def __init__(self, header, fun):
        self.header = header
        self.fun = fun
    
    def entry(self, element, auto_dir, directive):
        content = self.fun(element, auto_dir, directive)
        
        if isinstance(content, addnodes.pending_xref):
            return nodes.entry('', nodes.paragraph('', '', content))
        if isinstance(content, str):
            return nodes.entry('', nodes.paragraph(text=content))
        if isinstance(content, nodes.inline):
            return nodes.entry('', nodes.paragraph('', '', content))
        if isinstance(content, nodes.reference):
            return nodes.entry('', nodes.paragraph('', '', content))
        if isinstance(content, nodes.image):
            return nodes.entry('', content)
        if isinstance(content, nodes.paragraph):
            return nodes.entry('', content)
        if isinstance(content, nodes.entry):
            return content
        if content is None:
            return None
        raise TypeError(content)
    
    def colspec(self):
        return nodes.colspec(colwidth=10)
    
        
    def header_entry(self):
        return nodes.entry('', nodes.paragraph(text=self.header))

    
    


        

class MemberListTable:
    
    def __init__(self, header, types, *column_specs):
        self.header = header
        self.types = types
        self.column_specs = column_specs
   
    def sections(self, auto_dir):
        if isinstance(auto_dir.element, metamodel.Namespace):
            members = auto_dir.element.members
        elif isinstance(auto_dir.element, metamodel.Object):
            members = []
            
            if auto_dir.element.type.name == 'SymbolCatalogue':
            
                for prop in auto_dir.element.type.attributes.values():
                    if isinstance(prop, metamodel.CompositionProperty):
                        for child in prop._get_values_(auto_dir.element):
                            if child.name:
                                members.append(child)
        else:
            raise TypeError(auto_dir.element)

        row_entries = []
        non_empty_col_numbers = set()
        for member in members:
            if isinstance(member, self.types):
                entries = [col.entry(member, auto_dir, auto_dir) for col in self.column_specs]
                all_entries_none = True
                for col_nr, entry in enumerate(entries):
                    if entry is not None:
                        non_empty_col_numbers.add(col_nr)
                        all_entries_none = False
                if not all_entries_none:
                    row_entries.append(entries)
                    
                

                
                
                
        #       rows.append(nodes.row('', *[col.entry(member, auto_dir, auto_dir) for col in self.column_specs]))
                
                
                
        #  nodes.entry('', nodes.paragraph(text=''))

        if row_entries:
            
            section = nodes.section(ids=[f'member-list-section-{self.header}'])
            section.append(nodes.title(text=self.header))
            yield section
            
            non_empty_col_numbers = sorted(non_empty_col_numbers)
            
            non_empty_cols = [self.column_specs[col_nr] for col_nr in non_empty_col_numbers]

            tgroup = nodes.tgroup()
            tgroup.extend(col.colspec() for col in non_empty_cols)
            thead = nodes.thead('', nodes.row('', *[col.header_entry() for col in non_empty_cols]))
            tgroup.append(thead)
            
            rows = []
            for entries in row_entries:
                entries_in_non_empty_cols = []
                for col_nr in non_empty_col_numbers:
                    entry = entries[col_nr]
                    if entry is None:
                        entry = nodes.entry('', nodes.paragraph(text=''))
                    entries_in_non_empty_cols.append(entry)
                rows.append(nodes.row('', *entries_in_non_empty_cols))
            tbody = nodes.tbody('', *rows)
            tgroup.append(tbody)
            table = nodes.table(align="left")
            table.append(tgroup)
            yield table            
            
            
            
            
            
            
            
            
            '''

            tgroup = nodes.tgroup()
            for col_nr in sorted(non_empty_col_numbers):
                tgroup.extend(self.column_specs[col_nr].colspec())
                
            thead = nodes.thead('', nodes.row('', *[col.header_entry() for col in self.column_specs]))
                
''
            
            
            thead = nodes.thead('', nodes.row('', *[col.header_entry() for col in self.column_specs]))
            tgroup.append(thead)
            tbody = nodes.tbody('', *rows)
            tgroup.append(tbody)
            table = nodes.table(align="left")
            table.append(tgroup)
            yield table'''

         
NAME_COLUMN = ColumnSpec(
    'Name',
    lambda element, auto_dir, directive: element.name)



def _name_column_with_ref_target(element, auto_dir, directive):
    
    labelid = directive.sp._get_element_ref_name(element)
    
    
    entry = nodes.entry()
    
    #, nodes.paragraph('', '', content))
    
    
    
    
    entry.append(nodes.target(refid=labelid))
    
    
    
   
    paragraph = nodes.paragraph(ids=[labelid], names=[labelid])
    
    paragraph.append(nodes.inline(text=element.name))
    entry.append(paragraph)
    
    
    ref_name = labelid.lower()
    
    
    standard_domain = directive.env.domains.standard_domain
    standard_domain.anonlabels[ref_name] = directive.env.docname, labelid
    standard_domain.labels[ref_name] = directive.env.docname, labelid, element.name


    
    
    
    
    
    
    
    
    
    return entry
    


NAME_COLUMN_WITH_REF_TARGET = ColumnSpec(
    'Name',
    _name_column_with_ref_target)

NAME_COLUMN_WITH_REF = ColumnSpec(
    'Name',
    lambda element, auto_dir, directive: auto_dir.get_element_reference(element))

MULTIPLICITY_COLUMN = ColumnSpec(
    'Multiplicity',
    lambda element, auto_dir, directive: multiplicity_string(element.lower, element.upper))


def type_expression_column_text(element, auto_dir, directive):
    paragraph = nodes.paragraph()
    paragraph.extend(format_type_expression(element.type, directive))
    return paragraph


    
def format_type_expression(type_expression, directive):
    parts = []
    if isinstance(type_expression, (metamodel.Type, metamodel.TypeParameter)):
        parts.append(directive.get_element_reference(type_expression))
    elif isinstance(type_expression, metamodel.UnionType):
        for base in type_expression.bases:
            if parts:
                parts.append(nodes.inline(text=' | '))
            parts.extend(format_type_expression(base, directive))
    elif isinstance(type_expression, metamodel.BoundType):
        parts += [
            nodes.inline(text='<'),
            directive.get_element_reference(type_expression.base)]
        binding_parts = []
        for binding in type_expression.bindings:
            if binding_parts:
                binding_parts.append(nodes.inline(text=', '))
            else:
                binding_parts.append(nodes.inline(text=' with '))
            binding_parts += [
                directive.get_element_reference(binding.parameter),
                nodes.inline(text=' → ')]
            binding_parts += format_type_expression(binding.type, directive)
        parts += binding_parts
        parts.append(nodes.inline(text='>'))
    else:
        raise TypeError(type_expression)
        TODO
    return parts






def symbol_preview_column(element, auto_dir, directive):
    if not isinstance(element, metamodel.Object):
        return

    if element.type.name != 'Symbol':
        return
    
    variant = element.Variants[0]
    assert variant.VariantNumber == 0

    from dexpilib.io.svg.writer import ShapeToSvg, Box
    
    color_cls = directive.sp.model_set['Core'].search_unique('Color', metamodel.Type)

    white = color_cls(R=255, G=255, B=255)
    box = Box(variant.MinX, variant.MaxX, variant.MinY, variant.MaxY)
    box.expand(1)
    svg = ShapeToSvg(directive.sp.model_set, variant, box, white)
    xml = svg.svg.xml()
    
    xml.attrib['width'] = f'{1*float(xml.attrib["width"])}mm'
    xml.attrib['height'] = f'{1*float(xml.attrib["height"])}mm'

    diagram_name = f'symbol-preview-{element.name}.svg'
    out_dir = directive.env.app.outdir / '_images'
    out_dir.mkdir(parents=True, exist_ok=True)
    etree.ElementTree(xml).write(out_dir / diagram_name)

    return nodes.image(uri=str(out_dir / diagram_name))


SYMBOL_PREVIEW_COLUMN = ColumnSpec(
    'Preview',
    symbol_preview_column)
    
    
def symbol_usage_column(element, auto_dir, directive):
    if not isinstance(element, metamodel.Object):
        return

    if element.type.name != 'Symbol':
        return

    usage = directive.sp.metadata.get(element, 'usage')

    if not usage:
        return None

    return directive.get_element_reference(directive.sp.get_element(usage))



def extended_type_column(element, auto_dir, directive):
    baseType = getattr(element, 'baseType', None)
    if not baseType:
        return
    return directive.get_element_reference(baseType)


EXTENDED_TYPE_COLUMN = ColumnSpec(
    'Extended Class',
    extended_type_column)

SYMBOL_USAGE_COLUMN = ColumnSpec(
    'Usage',
    symbol_usage_column)

TYPE_EXPRESSION_COLUMN = ColumnSpec(
    'Type',
    type_expression_column_text)

TEMPLATE_TYPE_EXPRESSION_COLUMN = ColumnSpec(
    'Default Type',
    type_expression_column_text)

TYPE_NAME_COLUMN = ColumnSpec(
    'Type',
    lambda element, auto_dir, directive: auto_dir.sp.get_type_name(element))

ORDERED_COLUMN = ColumnSpec(
    'Ordered',
    lambda element, auto_dir, directive: 'yes' if element.isOrdered else 'no')

def _rdl_uri_column(element, auto_dir, directive):
    rdl_uri = directive.sp.metadata.get(element, 'rdl_uri', None)
    if not rdl_uri:
        return None
    return nodes.reference(refuri=rdl_uri, text=rdl_uri)

RDL_URI_COLUMN = ColumnSpec(
    'RDL URI',
    _rdl_uri_column)

def _uom_symbol_column(element, auto_dir, directive):
    return directive.sp.metadata.get(element, 'un_symbol', None) # TODO rename uom!!!

UOM_SYMBOL_COLUMN = ColumnSpec(
    'Symbol',
    _uom_symbol_column)

UN_CODE_COLUMN = ColumnSpec(
    'UN Code',
    lambda element, auto_dir, directive: auto_dir.sp.metadata.get(element, 'un_code', None))

def _rdl_label_column(element, auto_dir, directive):
    return directive.sp.metadata.get(element, 'rdl_label', None)

RDL_LABEL_COLUMN = ColumnSpec(
    'RDL Label',
    _rdl_label_column)

UNIQUE_COLUMN = ColumnSpec(
    'Unique',
    lambda element, auto_dir, directive: 'yes' if element.isUnique else 'no')

OPPOSITE_MULTIPLICITY_COLUMN = ColumnSpec(
    'Opposite Multiplicity',
    lambda element, auto_dir, directive: multiplicity_string(element.oppositeLower, element.oppositeUpper))

PREVIEW_TEXT_COLUMN = ColumnSpec(
    '',
    lambda element, auto_dir, directive: auto_dir.get_preview_text(element))


def _preview_text_column_html_only(element, auto_dir, directive):
    if auto_dir.sp.is_html:
        return auto_dir.get_preview_text(element)

PREVIEW_TEXT_COLUMN_HTML_ONLY = ColumnSpec(
    '',
    _preview_text_column_html_only)


MEMBER_LIST_TABLES = [
    MemberListTable(
        'Packages',
        metamodel.Package,
        NAME_COLUMN_WITH_REF,
        PREVIEW_TEXT_COLUMN_HTML_ONLY),
    MemberListTable(
        'Classes',
        metamodel.Class,
        NAME_COLUMN_WITH_REF,
        TYPE_NAME_COLUMN,
        PREVIEW_TEXT_COLUMN_HTML_ONLY),
    MemberListTable(
        'Class Extensions',
        metamodel.ClassExtension,
        NAME_COLUMN_WITH_REF,
        EXTENDED_TYPE_COLUMN),
    MemberListTable(
        'Data Types',
        metamodel.DataType,
        NAME_COLUMN_WITH_REF,
        TYPE_NAME_COLUMN,
        PREVIEW_TEXT_COLUMN_HTML_ONLY),
    MemberListTable(
        'Literals',
        metamodel.EnumerationLiteral,
        NAME_COLUMN_WITH_REF_TARGET,
        UOM_SYMBOL_COLUMN,
        UN_CODE_COLUMN,
        RDL_LABEL_COLUMN,
        RDL_URI_COLUMN),
    MemberListTable(
        'Objects',
        metamodel.Object,
        NAME_COLUMN_WITH_REF,
        TYPE_EXPRESSION_COLUMN,
        SYMBOL_PREVIEW_COLUMN,
        SYMBOL_USAGE_COLUMN),
    MemberListTable(
        'Type Parameters',
        metamodel.TypeParameter,
        NAME_COLUMN_WITH_REF,
        TEMPLATE_TYPE_EXPRESSION_COLUMN,
        PREVIEW_TEXT_COLUMN_HTML_ONLY),
    MemberListTable(
        'Composition Properties',
        metamodel.CompositionProperty,
        NAME_COLUMN_WITH_REF,
        MULTIPLICITY_COLUMN,
        TYPE_EXPRESSION_COLUMN,
        ORDERED_COLUMN,
        PREVIEW_TEXT_COLUMN_HTML_ONLY),
    MemberListTable(
        'Reference Properties',
        metamodel.ReferenceProperty,
        NAME_COLUMN_WITH_REF,
        MULTIPLICITY_COLUMN,
        TYPE_EXPRESSION_COLUMN,
        ORDERED_COLUMN,
        UNIQUE_COLUMN,
        OPPOSITE_MULTIPLICITY_COLUMN,
        PREVIEW_TEXT_COLUMN_HTML_ONLY),
    MemberListTable(
        'Data Properties',
        metamodel.DataProperty,
        NAME_COLUMN_WITH_REF,
        MULTIPLICITY_COLUMN,
        TYPE_EXPRESSION_COLUMN,
        ORDERED_COLUMN,
        UNIQUE_COLUMN,
        PREVIEW_TEXT_COLUMN_HTML_ONLY),
    ]


class InternalAutoelementDirective(SphinxDirective):

    option_spec = {'element': directives.unchanged}
    
    diagram_nr = 0
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        sp = self.env.domains.get('sp')
        assert isinstance(sp, SpeciPyDomain)
        self.sp = sp
        self.element_qname: str = self.options.get('element')
        self.element = sp.get_element(self.element_qname)

    def run(self) -> int:
        
        if self.element_qname.endswith('.CustomAttribute') or self.element_qname.endswith('.CustomObject'):
            REMOVE
            return []

        return list(itertools.chain(
            self.hidden_toc(),
            self.navigation_bar(),
            self.overview(),
            self.member_lists(),
            self.details(),
            self.members_details(),
            self.example()))
        
        
    def navigation_bar(self):
        owners = [self.element]
        while owners[-1].owner:
            owners.append(owners[-1].owner)
        owners.reverse()
        owners_paragraph = nodes.paragraph()
        for owner in owners:
            if owner is not owners[0]:
                owners_paragraph.append(nodes.inline(text='::'))
            if owner is self.element:
                owners_paragraph.append(nodes.inline(text=self.element.name))
            else:
                owners_paragraph.append(self.get_element_reference(owner))
        yield owners_paragraph
            

        
    def overview(self):
        
        description = self.sp.metadata.get(self.element, 'description', None)
        description_from_rdl = self.sp.metadata.get(self.element, 'description_from_rdl', None)
        if description_from_rdl:
            rdl_uri = self.sp.metadata.get(self.element, 'rdl_uri', None)
            assert rdl_uri
            description = description.strip()
            if description.endswith('.'):
                description = description[:-1]
                description = description + f' (from `{rdl_uri} <{rdl_uri}>`_).'

        return self.NEW_section(
            'Overview', [
                self.base_info(),
                self.rdl_paragraph_with_intro(self.element),
                self.nested_text(description),
                self.uml_diagram(),
                self.overview_tables()])


   
 
    def uml_diagram(self):
        
        
        if self.sp.is_html:
            return self.uml_diagram_html()
        else:
            assert self.sp.is_latex
            return self.uml_diagram_latex()
        
        
        
    def uml_diagram_html(self):
        

        if not isinstance(self.element, metamodel.Type):
            return []

 #       if self.element.name not in ['ConceptualModel', 'CentrifugalPump']:
 #           return []
        
        def get_link(element):
            
            if element is self.element:
                return None
            
            if target_doc_name := self.sp.doc_name_by_element.get(element):
                return f'../{target_doc_name}.html'
            else:
                # TODO: other cases?
                owner = element.owner
                target_doc_name = self.sp.doc_name_by_element.get(owner)
                return f'../{target_doc_name}.html#{self.sp._get_element_ref_name(element)}'

        diagram = uml_diagrams.make_diagram(
            [self.element],
            'svg',
            get_link=get_link,
            cache_dir=self.sp.cache_dir,
            usages_by_class=self.sp.usages_by_class)
        
        if diagram is None:
            return []

        InternalAutoelementDirective.diagram_nr += 1

        diagram_name = f'uml-{InternalAutoelementDirective.diagram_nr}.svg'

        out_dir = self.env.srcdir / 'auto_images'
        out_dir.mkdir(parents=True, exist_ok=True)
        out_path = out_dir / diagram_name
        assert not  out_path.is_file()
        etree.ElementTree(diagram).write(out_path)
        assert out_path.is_file()

        nesting_level = self.env.docname.count('/')
        refuri = nesting_level * '../' + '_images/' + diagram_name

        link_paragraph = nodes.paragraph()
        link_paragraph.extend([
            nodes.inline(text='('),
            nodes.reference(refuri=refuri, text='Fullscreen Diagram'),
            nodes.inline(text=')'),])

        return [
            link_paragraph,
            nodes.image(uri=f'/auto_images/{diagram_name}')]
        
        
    def uml_diagram_latex(self):


        if not isinstance(self.element, metamodel.Type):
            return []
        
        if self.element.name in [
                'NominalDiameterStandardClassification',
                'NominalPressureStandardClassification',
                'PureMaterialComponent']:
            return []

        diagram = uml_diagrams.make_diagram(
            [self.element],
            'tex',
            cache_dir=self.sp.cache_dir,
            usages_by_class=self.sp.usages_by_class)

        if diagram is None:
            return []

        InternalAutoelementDirective.diagram_nr += 1

        diagram_name = f'uml-{InternalAutoelementDirective.diagram_nr}.tex'

        out_dir = self.env.app.outdir
        out_dir.mkdir(parents=True, exist_ok=True)
        out_path = out_dir / diagram_name
        assert not  out_path.is_file()
        
        # TODO: encoding?
        out_path.write_text(diagram, encoding='utf-8')
        
        return [nodes.image(uri=f'/{diagram_name}')]
                            


            
    def get_preview_text(self, element):
        
        description = self.sp.metadata.get(element, 'description', None)
        parsed_description = self.nested_text(description)
        if parsed_description and isinstance(parsed_description[0], nodes.paragraph):
            return parsed_description[0]
           
        
    def details(self):
        
        return self.nested_sections(self.sp.metadata.get(self.element, 'details', None))

                
        
    def rdl_data(self):
        REMOVE
        rdl_uri = self.sp.metadata.get(self.element, 'rdl_uri', None)
        rdl_label = self.sp.metadata.get(self.element, 'rdl_label', None)
        
        if not rdl_uri and not rdl_label:
            return []
        
        label = f'{rdl_label} ({rdl_uri})'
        paragraph = nodes.paragraph(text='RDL: ')
        paragraph.append(nodes.reference(refuri=rdl_uri, text=label))
        return [paragraph]
        
        
    def nested_text(self, text):
        # TODO: check no sections
        if not (text and text.strip()):
            return []
       
        with self.sp.nested_rst(self.env.docname):
            return self.parse_text_to_nodes(text, allow_section_headings=True)
            
        
    def nested_sections(self, text):
        # TODO: check no sections
        if not (text and text.strip()):
            return []
       
        with self.sp.nested_rst(self.env.docname):
            return self.parse_text_to_nodes(text, allow_section_headings=True)
        
        

        
        
        
    def hidden_toc(self):
        
        #if self.element.name == 'SymbolCatalogue':
        #    a=1
        
        children_with_doc = self.sp.get_element_children_with_own_doc(self.element)
        if children_with_doc:
            docnames = [self.sp.doc_name_by_element[child] for child in children_with_doc]
            toctree = addnodes.toctree()
            toctree['includefiles'] = docnames
            toctree['entries'] = [(None, docname) for docname in docnames]
            toctree['glob'] = None
            toctree['hidden'] = True
            yield toctree

    def base_info(self):
        
        if isinstance(self.element, metamodel.Object):
            node = nodes.rubric()
            node.extend([
                nodes.inline(text='Object ('),
                self.get_element_reference(self.element.type),
                nodes.inline(text=')'),])
            yield node
        elif isinstance(self.element, metamodel.ClassExtension):
            node = nodes.rubric()
            node.extend([
                nodes.inline(text='Class Extension of '),
                self.get_element_reference(self.element.baseType)])
            yield node
        else:
        
            yield nodes.rubric(text=self.sp.get_type_name(self.element))
        if isinstance(self.element, metamodel.Model):
            yield nodes.paragraph(text=f'URI: {self.element.uri}')
            
            code_paragraph = nodes.paragraph()
            code_paragraph.append(
                nodes.reference(refuri=f'../../{self.element.name} RC 1.xml',
                                text=f'{self.element.name} RC 1.xml in DEXPI XML'))
            yield code_paragraph
 
                

    def section(self, header, content):
        if not isinstance(content, list):
            content = list(content)
        if content:
            section = nodes.section(ids=[header])
            section.append(nodes.title(text=header))
            section.extend(content)
            yield section
            
            
    def NEW_section(self, header, content):
        content = list(itertools.chain.from_iterable(content))
        if content:
            section = nodes.section(ids=[header])
            section.append(nodes.title(text=header))
            section.extend(content)
            yield section

            
    def overview_tables(self):
        
        if isinstance(self.element, metamodel.Type):
            if self.element.superTypes:
                yield nodes.rubric(text='Supertypes')
                list_ = nodes.bullet_list(bullet='-')
                for st in sorted(
                        self.element.superTypes,
                        key=lambda el: el.name):
                    para = nodes.paragraph()
                    para.append(self.get_element_reference(st))
                    li = nodes.list_item()
                    li.append(para)
                    list_.append(li)
                yield list_
                
            subTypes = self.element.subTypes
            if subTypes:
                yield nodes.rubric(text='Subtypes')
                list_ = nodes.bullet_list(bullet='-')
                for st in sorted(
                        subTypes,
                        key=lambda el: el.name):
                    
                    para = nodes.paragraph()
                    para.append(self.get_element_reference(st))
                    li = nodes.list_item()
                    li.append(para)
                    list_.append(li)
                yield list_
            
    def member_lists(self):

        yield from self.section(
            'Members',
            itertools.chain.from_iterable(table.sections(self) for table in MEMBER_LIST_TABLES))
   
    def get_element_reference(self, element, label=None):
        return self.sp.get_element_reference(element, label, ref_docname=self.env.docname)[0][0] # TODO
  
  
    def members_details(self):
        element = self.element
        if isinstance(element, metamodel.Type):
            for member in self.element:
                if isinstance(member, metamodel.Property):
                    yield from self.property_details(member)
                if isinstance(member, metamodel.TypeParameter):
                    yield from self.type_parameter_details(member)
        if isinstance(element, metamodel.ClassExtension):
            for member in self.element:
                if isinstance(member, metamodel.Property):
                    yield from self.property_details(member)
                    
                    
    def example(self):
        element = self.element
        if isinstance(element, metamodel.ConcreteClass):
            constructor = element
        else:
            constructor = None # TODO
            
        if constructor:
            
            constructor_kwargs = {}
            
            for prop in self.element.ownedAttributes:
                prop_example_value = self.sp.metadata.get(prop, 'exampleValue', None)
                if prop_example_value is None:
                    continue
                constructor_kwargs[prop.name] = prop_example_value

            example_value = constructor(**constructor_kwargs)
            
            
            model = metamodel.Model(name='example', uri='http://www.example.org')
            model.add(example_value)
            model_xml = XmlExporter(model).xml
            example_xml = None
            for child in model_xml:
                if child.tag != 'Import':
                    if example_xml is not None:
                        RAISE
                    example_xml = child
            if example_xml is None:
                RAISE
            xml_code = etree.tostring(example_xml, pretty_print=True).decode('utf-8')
            xml_code = textwrap.indent(xml_code, '   ')
            rst_code = f'.. code-block:: xml\n\n{xml_code}'
            section = nodes.section(ids=[f'{self.element.qualifiedName_chained}-example'])
            section.append(nodes.title(text='Example in DEXPI XML'))
            section.extend(self.parse_text_to_nodes(rst_code))

            yield section 
            
            
                    
                    
    def paragraph_with_intro(self, intro, content):
        paragraph = nodes.paragraph()
        paragraph.extend([
            nodes.strong(text=f'{intro}:'),
            nodes.inline(text=' ')])
        
        if isinstance(content, str):
            content = nodes.inline(text=content)
        if isinstance(content, nodes.Node):
            content = [content]
        paragraph.extend(content)
        return paragraph
    
    
    def rdl_paragraph_with_intro(self, element):

        
        rdl_uri = self.sp.metadata.get(element, 'rdl_uri', None)
        rdl_label = self.sp.metadata.get(element, 'rdl_label', None)
        if not rdl_uri and not rdl_label:
            return []
        if not (rdl_uri and rdl_label):
            RAISE
            
            
        if self.sp.is_html:
            label = f'{rdl_label} ({rdl_uri})'
            return [self.paragraph_with_intro('RDL', nodes.reference(refuri=rdl_uri, text=label))]
        else:
            return [
                self.paragraph_with_intro('RDL', nodes.reference(refuri=rdl_uri, text=rdl_uri)),
                nodes.paragraph(text=rdl_label)]
            
        


            
            
    def property_details(self, prop):
        section = self.sp.create_element_section(prop, self.env.docname)
        section.append(nodes.rubric(text=self.sp.get_type_name(prop)))
        section.extend(self.nested_text(self.sp.metadata.get(prop, 'description', None)))

        section.append(
            self.paragraph_with_intro(
                'Type',
                format_type_expression(prop.type, self)))
        section.append(
            self.paragraph_with_intro(
                'Multiplicity',
                multiplicity_string(prop.lower, prop.upper)))
        section.append(
            self.paragraph_with_intro(
                'Ordered',
                'yes' if prop.isOrdered else 'no'))
        if isinstance(prop, (metamodel.DataProperty, metamodel.ReferenceProperty)):
            section.append(
                self.paragraph_with_intro(
                    'Unique',
                    'yes' if prop.isUnique else 'no'))
        if isinstance(prop, metamodel.ReferenceProperty):
            section.append(
                self.paragraph_with_intro(
                    'Opposite Multiplicity',
                    multiplicity_string(prop.oppositeLower, prop.oppositeUpper)))
        section.extend(self.rdl_paragraph_with_intro(prop))

        section.extend(self.nested_sections(self.sp.metadata.get(prop, 'details', None)))
        
        
        example_value = self.sp.metadata.get(prop, 'exampleValue', None)
        if example_value is not None:
            example_section = nodes.section(ids=[f'{self.sp._get_element_ref_name(prop)}-example'])
            section.append(example_section)
            example_section.append(nodes.title(text='Example'))
            
            
            if isinstance(example_value, list):
                assert prop.lower <= len(example_value)
                assert prop.upper is None or prop.upper >= len(example_value)
                
                
                paragraph = nodes.paragraph(classes=['sp-dense'])
                paragraph.append(nodes.inline(text='['))
                example_section.append(paragraph)
                
                for nr, value in enumerate(example_value):
                    is_last_value = nr + 1 == len(example_value)
                    example_section.extend(self.value_representation_paragraphs(value, 1, True))
                    if not is_last_value:
                        example_section[-1].append(nodes.inline(text=','))
                        
                paragraph = nodes.paragraph(classes=['sp-dense'])
                paragraph.append(nodes.inline(text=']'))
                example_section.append(paragraph)
                

                

            else:
                assert prop.lower in [0, 1]
                assert prop.upper is None or prop.upper >= 1
            
                example_section.extend(self.value_representation_paragraphs(example_value))
        yield section

    def type_parameter_details(self, type_parameter):
        section = self.sp.create_element_section(type_parameter, self.env.docname)
        section.append(nodes.rubric(text=self.sp.get_type_name(type_parameter)))
        section.extend(self.nested_text(self.sp.metadata.get(type_parameter, 'description', None)))
        section.append(
            self.paragraph_with_intro(
                'Default Type',
                format_type_expression(type_parameter.type, self)))
        section.extend(self.nested_sections(self.sp.metadata.get(type_parameter, 'details', None)))
        yield section

        
    def member_groups(self, group_infos):
        for header, members, types, data in group_infos:
            members_in_group = [member for member in members if isinstance(member, types)]
            if members_in_group:
                yield header, members_in_group, data
                
                
                
    def value_representation_paragraphs(self, value, indent=0, indent_first=False):
        
        nbsp_per_indent = 5

        paragraphs = []
        
        def new_paragraph():
            paragraph = nodes.paragraph(classes=['sp-dense'])
            paragraph.append(nodes.inline(text=indent*nbsp_per_indent*NBSP))
            paragraphs.append(paragraph)
            return paragraph
        
        if indent_first:
            paragraph = new_paragraph()
        else:
            paragraph = nodes.paragraph(classes=['sp-dense'])
            paragraphs.append(paragraph)

        indent += 1

        if isinstance(value, str):
            paragraph.append(nodes.literal(text=f'"{value}"'))
        elif isinstance(value, bool):
            paragraph.append(nodes.literal(text=str(value).lower()))
        elif isinstance(value, int):
            paragraph.append(nodes.literal(text=str(value)))
        elif isinstance(value, float):
            paragraph.append(nodes.literal(text=str(value)))
        elif isinstance(value,datetime.datetime):
            paragraph.append(nodes.literal(text=value.isoformat()))
        elif isinstance(value, metamodel.EnumerationLiteral):
            paragraph.append(self.get_element_reference(value))
        elif isinstance(value, (metamodel.AggregatedDataValue, metamodel.Object)):
            paragraph.append(self.get_element_reference(value.type))
            paragraph.append(nodes.inline(text='('))
            for prop_nr, (prop, prop_values) in enumerate(
                    sorted(value._attribute_values_.items(),
                           key=lambda prop_and_values: prop_and_values[0].name)):
                is_last_prop = prop_nr + 1 == len(value._attribute_values_)

                paragraph = new_paragraph()
                paragraph.append(self.get_element_reference(prop))
                paragraph.append(nodes.inline(text=' = '))
                
                assert len(prop_values) >= prop.lower
                assert prop.upper is None or len(prop_values) <= prop.upper
                
                if prop.upper == 1:
                    prop_value_paragraphs = self.value_representation_paragraphs(prop_values[0], indent)
                    paragraph.extend(prop_value_paragraphs[0])
                    paragraphs.extend(prop_value_paragraphs[1:])
                    paragraph = paragraphs[-1]
                else:
                    paragraph.append(nodes.inline(text='['))
                    indent += 1
                    for nr, prop_value in enumerate(prop_values):
                        is_last_prop_value = nr + 1 == len(prop_values)
                        prop_value_paragraphs = self.value_representation_paragraphs(prop_value, indent, True)
                        paragraphs.extend(prop_value_paragraphs)
                        paragraph = paragraphs[-1]
                        if not is_last_prop_value:
                            paragraph.append(nodes.inline(text=','))
                    paragraph.append(nodes.inline(text=']'))
                    indent -= 1
                if not is_last_prop:
                    paragraph.append(nodes.inline(text=','))
            paragraph.append(nodes.inline(text=')'))
        else:
            paragraph.append(nodes.literal(text='mimi '+repr(value)))
        return paragraphs
            
        
        
                
                
            
            
        
        
        
    
            
            
        


class SpeciPyDomain(Domain):

    name = 'sp'
    label = 'SpeciPy'
    directives = {
        'internal-autoelement': InternalAutoelementDirective,
    }
    
    roles = {
        'element': ElementRole(),
        'name': NameRole(),
    }

    def __init__(self, env: BuildEnvironment) -> None:
        super().__init__(env)
        

        self._specificator_data: None | SpecificatorData = None

        self.doc_name_by_element: dict[metamodel.Model, str] = {}
        
        
        self._nested_docnames: list[str] = []
        
        self._usages_by_class = None
        self.usages_by_class

        
    @property
    def is_latex(self):
        return self.env.app.builder.name == 'latex'
    
    @property
    def is_html(self):
        return self.env.app.builder.name == 'html'
        
        

    @property
    def usages_by_class(self):
        usages_by_class = self._usages_by_class
        if usages_by_class is None:
            usages_by_class = {}
            
            namespaces_to_handle = set(self.model_set)
            handled_namespaces = set()
            while namespaces_to_handle:
                member = namespaces_to_handle.pop()
                
                if isinstance(member, metamodel.Namespace):
                    new_namespaces = set(member)
                    new_namespaces.difference_update(handled_namespaces)
                    namespaces_to_handle.update(new_namespaces)
                    
                if isinstance(member, metamodel.ObjectProperty):
                    prop_type = member.type

                    if isinstance(prop_type, metamodel.Class):
                        pass
                    elif isinstance(prop_type, metamodel.BoundClass):
                        prop_type = prop_type.base
                        assert isinstance(prop_type, metamodel.Class)
                    else:
                        raise Exception(type(prop_type))
                    usages_by_class.setdefault(prop_type, []).append(member)


            
            self._usages_by_class = usages_by_class
        return usages_by_class
            
            
        
        
    @contextlib.contextmanager
    def nested_rst(self, docname: str):
        self._nested_docnames.append(docname)
        yield
        self._nested_docnames.pop()
        
        
        
        
    @property
    def metadata(self) -> MetaData:
        assert len(self.model_set.metadata_by_name) == 1
        return list(self.model_set.metadata_by_name.items())[0][1]

    @property
    def cache_dir(self) -> Path | None:
        cache_dir = self._specificator_data.config.get('cache_dir')
        if cache_dir:
            return Path(str(cache_dir))
        else:
            return None

    @property
    def model_set(self) -> metamodel.ModelSet:
        return self.specificator_data.model_set

    @property
    def specificator_data(self) -> SpecificatorData:
        specificator_data = self._specificator_data
        if specificator_data is None:
            if not _DATA:
                raise Exception()
            specificator_data = self._specificator_data = _DATA
        return specificator_data
    
    #@property
    
    
    def get_element(self, qname: str):

        model_name, *sub_names = qname.split('.')
        element = self.model_set.get(model_name)
        if not element:
            return None
            TODO
        for name in sub_names:
            if type(element).__name__ == 'Object':
                child = None
                
                for prop in element.type.attributes.values():
                    if isinstance(prop, metamodel.CompositionProperty):
                        for child_ in prop._get_values_(element):
                            if child_.name == name:
                                child = child_
                                break
            else:
                try:
                    child = element.get(name)
                except Exception:
                    return None

            if not child:
                return None
                raise Exception(f'{element} has no member "{name}"')
                TODO
            element = child
        return element
    
    def create_automodel_documents(self):
        for doc_name in self.env.found_docs:
            doc_path = self.env.doc2path(doc_name)
            doc_raw = doc_path.read_text()
            for model_name in find_model_names_in_toc_trees(doc_raw):
                model = self.model_set.get(model_name)
                if not model:
                    continue
                    raise Exception(model_name)
                    TODO
                if model in self.doc_name_by_element:
                    TODO
                self._create_element_stub(model, doc_path.parent)
                
                
    def element_can_have_subdocs(self, element: metamodel.Element) -> bool:
        return isinstance(element, (metamodel.Model, metamodel.Package)) #, metamodel.Object))
                          
    def get_element_children_with_own_doc(self, element: metamodel.Element) -> list[metamodel.Element]:
        if isinstance(element, (metamodel.Model, metamodel.Package)):
            return [e for e in element.packagedElements if not isinstance(e, metamodel.Object)]
        return []
        
        if isinstance(element, (metamodel.Object)):
            
            children = []

            if element.type.name in ['SymbolCatalogue']:
                for prop in element.type.attributes.values():
                    if isinstance(prop, metamodel.CompositionProperty):
                        for child in prop._get_values_(element):
                            if child.name:
                                children.append(child)
            return children
                    

        return []
    
    
    def _get_element_ref_name(self, element):
        assert element not in self.doc_name_by_element
        # TODO: ambiguous refs due to 'lower'
        return f'sp:{element.qualifiedName}'.lower()
    
    
    def create_element_section(self, element, docname):
        
        labelid = self._get_element_ref_name(element)
        section = nodes.section(ids=[labelid])
        
        section.append(nodes.title(text=element.name))
        
        ref_name = self._get_element_ref_name(element).lower()

        standard_domain = self.env.domains.standard_domain
        standard_domain.anonlabels[ref_name] = docname, labelid
        standard_domain.labels[ref_name] = docname, labelid, element.name
        return section

        
          
    
    
    def get_element_reference(self, element, label=None, ref_docpath=None, ref_docname=None):
        
        if label is None:
            label = element.name
            
        nested_docname = self._nested_docnames[-1] if self._nested_docnames else None
        
        if nested_docname:
            assert ref_docname is None
            ref_docname = nested_docname
           # TODO: doc path for messages

        else:
            assert None in (ref_docpath, ref_docname)
            if ref_docname is None:
                assert ref_docpath
                ref_docname = self.env.path2doc(ref_docpath)

        if target_doc_name := self.doc_name_by_element.get(element):
            inline = nodes.inline(text=label)
            pending_xref = addnodes.pending_xref(
                classes=['sp-element'],
                refdoc=ref_docname,
                refdomain='std',
                refexplicit=True,
                reftarget=f'/{target_doc_name}',
                reftype='doc',
                refwarn=True)
            pending_xref.append(inline)
            return [pending_xref], []
        else:
            inline = nodes.inline(text=label)
            pending_xref = addnodes.pending_xref(
                classes=['sp-element'],
                refdoc=ref_docname,
                refdomain='std',
                refexplicit=True,
                reftarget=self._get_element_ref_name(element).lower(),
                reftype='ref',
                refwarn=True)
            pending_xref.append(inline)
            return [pending_xref], []

                
                
    


    def _create_element_stub(self, element, parent_dir):
        
        if type(element).__name__ == 'Object':
            return
        
        if self.element_can_have_subdocs(element):
            element_dir = parent_dir / element.name
            # TODO: check existing dir at least for top-level model
            element_dir.mkdir(parents=True, exist_ok=True)
            element_path = element_dir / f'{element.get_meta_class_name().lower()}-index.rst'
            for child in self.get_element_children_with_own_doc(element):
                self._create_element_stub(child, element_dir)
        else:
            
            if type(element).__name__ == 'Object':
                a=1
            
            element_path = parent_dir / f'{element.name}.rst'
        header = '\n'.join((
            element.name,
            len(element.name) * '=',
            '',
            '.. sp:internal-autoelement::',
            f'   :element: {element.qualifiedName_chained}'))
        element_path.write_text(header)
        self.doc_name_by_element[element] = self.env.path2doc(element_path)

    def get_type_name(self, type_or_element: metamodel.Element | T.Type[metamodel.Element]) -> str:
        if isinstance(type_or_element, metamodel.Element):
            type_ = type(type_or_element)
        else:
            type_ = type_or_element
        return dict(
            AggregatedDataType='Aggregated Data Type',
            Model='Model',
            Package='Package',
            AbstractClass='Abstract Class',
            ConcreteClass='Concrete Class',
            ClassExtension='Class Extension',
            DataProperty='Data Property',
            CompositionProperty='Composition Property',
            ReferenceProperty='Reference Property',
            DataTypeParameter='Data Type Parameter',
            ClassParameter='Class Parameter').get(type_.__name__, type_.__name__)
            
            
            
def on_build_finished(app, exception):
    shutil.copy(SP_CSS_PATH, app.outdir / '_static')
        
    

    


def setup(app: Sphinx) -> ExtensionMetadata:
    app.connect('builder-inited', create_automodel_documents)
    app.connect('build-finished', on_build_finished)
    app.add_directive('toctree', SpeciPyTocTree, override=True)
    app.add_domain(SpeciPyDomain)
    app.add_css_file('sp.css')

    # TODO: how to handle version; same as specificator?
    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


_MODEL_BY_NAME: None | dict[str, metamodel.Model] = None

from pnb.mcl.metamodel.standard import MetaData


_DATA = None

class SpecificatorData:
    
    def __init__(self, model_set, config):
        self.model_set = model_set
        self.config = config
        
    

_DATA: T.Optional[SpecificatorData] = None



@contextlib.contextmanager
def set_data(data: SpecificatorData):
    global _DATA
    if _DATA is not None:
        raise Exception('specificator data already set')
    _DATA = data
    
    yield
    
    _DATA = None
    
    

    
    
    








import pathlib

from tomlkit import table
from lxml import etree

from dexpi.specificator.formatter import Formatter

from pnb.mcl.metamodel import standard as metamodel
from pnb.mcl.io.xml import XmlExporter


def get_default_config():
    config = table()
    config.add('pretty_print', True)
    return config

def make(model_set, config):
    out_dir = pathlib.Path(config['out_dir'])
    out_dir.mkdir(parents=True, exist_ok=True)
    
    for model in model_set:
        exporter = XmlExporter(
            model,
            model_set)
        xml = exporter.xml
        out_path = out_dir / f'{model.name}.xml'
        etree.ElementTree(xml).write(
            out_path, encoding='utf-8', pretty_print=config['pretty_print'])

FORMATTER = Formatter(
    'xml', get_default_config, make)



from tomlkit import table
from dexpi.specificator.formatter import Formatter

from pnb.mcl.metamodel import standard as metamodel

Config = metamodel.AggregatedDataType('RdfConfiguration')
Config.add(metamodel.DataProperty('metamodel_namespace', str))




def get_default_config():
    config = table()
    config.add('metamodel_namespace', 'http://www.dexpi.org/temp/metamodel/')
    return config

def make(model_by_name, metadata, out_dir, config):
    from pnb.mcl.io.rdf import RdfExporter
    for name, model in model_by_name.items():
        graph = RdfExporter(model, metamodel_namespace=config['metamodel_namespace']).graph
        out_path = out_dir / f'{name}.ttl'
        out_path.parent.mkdir(parents=True, exist_ok=True)
        graph.serialize(out_path, format='ttl')

FORMATTER = Formatter(
    'rdf', get_default_config, make)



from sphinx.directives.other import *
from sphinx.directives.other import __


class SpeciPyTocTree(TocTree):
    """
    Extension of the standard Sphinx TocTree that can handle references to documentation pages
    generated by SpeciPy.
    """

    def parse_content(self, toctree: addnodes.toctree) -> None:
        """
        Populate ``toctree['entries']`` and ``toctree['includefiles']`` from content.
        """
        generated_docnames = frozenset(StandardDomain._virtual_doc_names)
        suffixes = self.config.source_suffix
        current_docname = self.env.docname
        glob = toctree['glob']

        # glob target documents
        all_docnames = self.env.found_docs.copy() | generated_docnames
        all_docnames.remove(current_docname)  # remove current document
        frozen_all_docnames = frozenset(all_docnames)

        excluded = Matcher(self.config.exclude_patterns)
        for entry in self.content:
            if not entry:
                continue
            # BEGIN MOD
            entry = self.handle_specipy_entry(entry)
            # END MOD
            # look for explicit titles ("Some Title <document>")
            explicit = explicit_title_re.match(entry)
            url_match = url_re.match(entry) is not None
            if glob and glob_re.match(entry) and not explicit and not url_match:
                pat_name = docname_join(current_docname, entry)
                doc_names = sorted(
                    docname
                    for docname in patfilter(all_docnames, pat_name)
                    # don't include generated documents in globs
                    if docname not in generated_docnames
                )
                if not doc_names:
                    logger.warning(
                        __("toctree glob pattern %r didn't match any documents"),
                        entry,
                        location=toctree,
                    )

                for docname in doc_names:
                    all_docnames.remove(docname)  # don't include it again
                    toctree['entries'].append((None, docname))
                    toctree['includefiles'].append(docname)
                continue

            if explicit:
                ref = explicit.group(2)
                title = explicit.group(1)
                docname = ref
            else:
                ref = docname = entry
                title = None

            # remove suffixes (backwards compatibility)
            for suffix in suffixes:
                if docname.endswith(suffix):
                    docname = docname.removesuffix(suffix)
                    break

            # absolutise filenames
            docname = docname_join(current_docname, docname)
            if url_match or ref == 'self':
                toctree['entries'].append((title, ref))
                continue

            if docname not in frozen_all_docnames:
                if excluded(str(self.env.doc2path(docname, False))):
                    msg = __('toctree contains reference to excluded document %r')
                    subtype = 'excluded'
                else:
                    msg = __('toctree contains reference to nonexisting document %r')
                    subtype = 'not_readable'

                logger.warning(
                    msg, docname, type='toc', subtype=subtype, location=toctree
                )
                self.env.note_reread()
                continue

            if docname in all_docnames:
                all_docnames.remove(docname)
            else:
                logger.warning(
                    __('duplicated entry found in toctree: %s'),
                    docname,
                    location=toctree,
                )

            toctree['entries'].append((title, docname))
            toctree['includefiles'].append(docname)

        # entries contains all entries (self references, external links etc.)
        if 'reversed' in self.options:
            toctree['entries'] = list(reversed(toctree['entries']))
            toctree['includefiles'] = list(reversed(toctree['includefiles']))


    def handle_specipy_entry(self, entry):
        match_obj = AUTOMODEL_PATTERN_FOR_TOC_ENTRY.match(entry)
        if not match_obj:
            return entry
        model_name = match_obj.group('model_name')
        specipy_domain = self.env.domains.get('sp')
        if not specipy_domain:
            # TODO: warn?
            return entry
        model = specipy_domain.model_set.get(model_name)
        if not model:
            # TODO: warn?
            return entry
        return '/' + specipy_domain.doc_name_by_element[model]


BASE_AUTOMODEL_RE = r'<#AUTOMODEL (?P<model_name>[a-zA-Z_][a-zA-Z_0-9]*)#>'

AUTOMODEL_PATTERN_FOR_TOC_ENTRY = re.compile(fr'^{BASE_AUTOMODEL_RE}$')
AUTOMODEL_PATTERN_FOR_DOC = re.compile(fr'^[ \t]*{BASE_AUTOMODEL_RE}[ \t]*$', flags=re.MULTILINE)

def find_model_names_in_toc_trees(raw_doc: str) -> list[str]:
    # TODO: make safer (e.g., check if pattern really appears in a toc tree)
    return [match_obj.group('model_name')
            for match_obj in AUTOMODEL_PATTERN_FOR_DOC.finditer(raw_doc)]



import logging
import textwrap
import typing as T
import unittest

from lxml import etree
from pnb.mcl.io.xml import XmlExporter

from dexpi.specificator import dsl_reader

DEFAULT_LOGGING_LEVEL = logging.getLogger().level
TEST_LOGGING_LEVEL = logging.WARNING + 1

DEFAULT_DOCUMENTATION_URL = dsl_reader.DOCUMENTATION_URL
TEST_DOCUMENTATION_URL = 'https://www.test.org/'


class TestDslReader(unittest.TestCase):

    def assert_models(self,
            src: str,
            expected_model_xml_by_name: T.Optional[dict[str, str]]=None):
        self.assert_messages_and_models(src, '', expected_model_xml_by_name)

    def assert_messages_and_models(self,
            src: str,
            expected_messages: str,
            expected_model_xml_by_name: T.Optional[dict[str, str]]=None):

        assert logging.getLogger().level == DEFAULT_LOGGING_LEVEL
        logging.getLogger().setLevel(TEST_LOGGING_LEVEL)
        try:
            reader = dsl_reader.DslReader(textwrap.dedent(src).strip())
        finally:
            logging.getLogger().setLevel(DEFAULT_LOGGING_LEVEL)

        assert dsl_reader.DOCUMENTATION_URL == DEFAULT_DOCUMENTATION_URL
        dsl_reader.DOCUMENTATION_URL = TEST_DOCUMENTATION_URL
        try:
            actual_messages = '\n\n'.join(str(msg) for msg in reader.messages)
        finally:
            dsl_reader.DOCUMENTATION_URL = DEFAULT_DOCUMENTATION_URL

        expected_messages = textwrap.dedent(expected_messages).strip()

        test_report_lines = []

        if actual_messages != expected_messages:
            test_report_lines += [
                '',
                '',
                'actual_messages do not match expected_messages',
                '----------------------------------------------',
                '',
                'actual_messages:',
                '',
                actual_messages,
                '',
                'expected_messages:',
                '',
                expected_messages]

        built_model_by_name = dict(reader.model_by_name)
        del built_model_by_name['Builtin']

        expected_model_xml_by_name = expected_model_xml_by_name or {}

        names = set(built_model_by_name)
        names.update(expected_model_xml_by_name)

        for name in sorted(names):

            built_model = built_model_by_name.get(name)
            if built_model is None:
                test_report_lines += [
                    '',
                    '',
                    f'expected model "{name}" was not built',
                    '-------------------------------' + len(name)*'-']
                continue
            built_model_xml = etree.tostring(
                XmlExporter(built_model).xml,
                pretty_print=True,
                encoding='unicode').strip()

            expected_model_xml = expected_model_xml_by_name.get(name)
            if expected_model_xml is None:
                test_report_lines += [
                    '',
                    '',
                    f'unexpected model "{name}" was built',
                    '------------------------------' + len(name)*'-',
                    '',
                    built_model_xml]
                continue
            expected_model_xml = textwrap.dedent(expected_model_xml).strip()

            if expected_model_xml != built_model_xml:
                test_report_lines += [
                    '',
                    '',
                    f'XML for model "{name}" does not match expected XML',
                    '--------------------------------------------' + len(name)*'-',
                    '',
                    'actual XML:',
                    '',
                    built_model_xml,
                    '',
                    'expected XML:',
                    '',
                    expected_model_xml]

        if test_report_lines:
            self.fail('\n'.join(test_report_lines))


    def test_self_messages(self):
        with self.assertRaises(AssertionError) as ctx:
            self.assert_messages_and_models('', 'spam')
        self.assertEqual(
            str(ctx.exception).strip(),
            textwrap.dedent('''
                actual_messages do not match expected_messages
                ----------------------------------------------
                
                actual_messages:
                
                
                
                expected_messages:
                
                spam''').strip())


    def test_fatal_error(self):
        self.assert_messages_and_models('''
                Model = MODEL(name='Model', uri='http://www.example.org/Model')
                Model.Car = CONCRETE_CLASS()
                
                def calculate_upper_limit_for_color():
                    return 1 + 1 + 1/0
                
                Model.Car.Color = DATA_PROPERTY(
                    lower=0,
                    upper=calculate_upper_limit_for_color(),
                    type=BUILTIN.String)
            ''', '''
                #1 --- FATAL ERROR: fatal-error ---
                An error has occurred while reading the DSL source files:
                
                  Traceback (most recent call last):
                    File "<string>", line 9, in <module>
                    File "<string>", line 5, in calculate_upper_limit_for_color
                  ZeroDivisionError: division by zero
                
                This error prevents any further processing of the source files.
                (see https://www.test.org/fatal-error)
            ''')


    def test_missing_parameter_name(self):
        self.assert_messages_and_models('''
                MODEL()
            ''', '''
                #1 --- ERROR: missing-parameter ---
                - File "<string>", line 1
                  MODEL
                Parameter "name" is required, but missing.
                (see https://www.test.org/missing-parameter)
            ''')


    def test_invalid_parameter_type_name(self):
        self.assert_messages_and_models('''
                MODEL(name=1)
            ''', '''
                #1 --- ERROR: invalid-parameter-type ---
                - File "<string>", line 1
                  MODEL "1"
                Parameter "name": got 1 of type int, but value type must be one of str, NoneType.
                (see https://www.test.org/invalid-parameter-type)
            ''')


    def test_no_valid_identifier_as_parameter_spaces(self):
        self.assert_messages_and_models('''
                MODEL(name='spaces not allowed')
            ''', '''
                #1 --- ERROR: no-valid-identifier ---
                - File "<string>", line 1
                  MODEL "spaces not allowed"
                Parameter "name": 'spaces not allowed' is no valid value: An identifier must start with [a-zA_Z], followed by an arbitrary number of [a-zA_Z0-9_].
                (see https://www.test.org/no-valid-identifier)
            ''')


    def test_no_valid_identifier_as_parameter_leading_underscore(self):
        self.assert_messages_and_models('''
                TEMPLATE_MODEL(name='_LeadingUnderscoreNotAllowed')
            ''', '''
                #1 --- ERROR: no-valid-identifier ---
                - File "<string>", line 1
                  TEMPLATE_MODEL "_LeadingUnderscoreNotAllowed"
                Parameter "name": '_LeadingUnderscoreNotAllowed' is no valid value: An identifier must start with [a-zA_Z], followed by an arbitrary number of [a-zA_Z0-9_].
                (see https://www.test.org/no-valid-identifier)
            ''')

    # TODO   
    # def test_no_valid_identifier_as_parameter_leading_underscore_2(self):
    #     self.assert_messages_and_models('''
    #             Model = MODEL(name='Model', uri='http://www.example.org/Model')
    #             Model._A = CONCRETE_CLASS()
    #         ''', '''
    #             #1 --- ERROR: no-valid-identifier ---
    #             - File "<string>", line 1
    #               TEMPLATE_MODEL "_LeadingUnderscoreNotAllowed"
    #             Parameter "name": '_LeadingUnderscoreNotAllowed' is no valid value: An identifier must start with [a-zA_Z], followed by an arbitrary number of [a-zA_Z0-9_].
    #             (see https://www.test.org/no-valid-identifier)
    #         ''')


    def test_no_valid_identifier_by_assignment_umlaut(self):
        self.assert_messages_and_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.Päckchen = PACKAGE()
            ''', '''
                #1 --- ERROR: no-valid-identifier ---
                - File "<string>", line 2
                  PACKAGE "Päckchen"
                Parameter "name": 'Päckchen' is no valid value: An identifier must start with [a-zA_Z], followed by an arbitrary number of [a-zA_Z0-9_].
                (see https://www.test.org/no-valid-identifier)
            ''', dict(
                Model='<Model name="Model" uri="http://www.ex.org"/>'))

    # TODO
    # def test_name_invalid_Package_2(self):
    #     self.assert_messages_and_models('''
    #             Model1 = TEMPLATE_MODEL(name='Model1')
    #             Model1._Package1 = PACKAGE()
    #         ''', '''
    #         ''')

    def test_no_owner(self):
        self.assert_messages_and_models('''
                CONCRETE_CLASS()
            ''', '''
                #1 --- ERROR: no-owner ---
                - File "<string>", line 1
                  CONCRETE_CLASS
                The element is skipped because it is not assigned to any owner.
                (see https://www.test.org/no-owner)
            ''')

    def test_owner_skipped(self):
        self.assert_messages_and_models('''
                A_PACKAGE = PACKAGE()
                A_PACKAGE.A_SUB_PACKAGE = PACKAGE()
            ''', '''
                #1 --- ERROR: no-owner ---
                - File "<string>", line 1
                  PACKAGE
                The element is skipped because it is not assigned to any owner.
                (see https://www.test.org/no-owner)
                
                #2 --- ERROR: owner-skipped ---
                - File "<string>", line 2
                  PACKAGE "A_SUB_PACKAGE"
                causes: #1
                The element is skipped because its owner is skipped.
                (see https://www.test.org/owner-skipped)
            ''')

    def test_owner_skipped_model(self):
        self.assert_messages_and_models('''
                A_MODEL = MODEL()
                A_MODEL.A_PACKAGE = PACKAGE()
            ''', '''
                #1 --- ERROR: missing-parameter ---
                - File "<string>", line 1
                  MODEL
                Parameter "name" is required, but missing.
                (see https://www.test.org/missing-parameter)
                
                #2 --- ERROR: owner-skipped ---
                - File "<string>", line 2
                  PACKAGE "A_PACKAGE"
                causes: #1
                The element is skipped because its owner is skipped.
                (see https://www.test.org/owner-skipped)
            ''')

    def test_owner_skipped_anonymous(self):
        self.assert_messages_and_models('''
                A_MODEL = MODEL()
                A_MODEL.A_SUB_PACKAGE.A_PACKAGE = PACKAGE()
            ''', '''
                #1 --- ERROR: missing-parameter ---
                - File "<string>", line 1
                  MODEL
                Parameter "name" is required, but missing.
                (see https://www.test.org/missing-parameter)
                
                #2 --- ERROR: owner-skipped ---
                - File "<string>", line 2
                  reference to "A_SUB_PACKAGE"
                causes: #1
                The element is skipped because its owner is skipped.
                (see https://www.test.org/owner-skipped)
                
                #3 --- ERROR: owner-skipped ---
                - File "<string>", line 2
                  PACKAGE "A_PACKAGE"
                causes: #2
                The element is skipped because its owner is skipped.
                (see https://www.test.org/owner-skipped)
            ''')

    def test_missing_metatype(self):
        self.assert_messages_and_models('''
                A = MODEL(name='Model', uri='http://www.ex.org')
                A.B
            ''', '''
                #1 --- ERROR: missing-metatype ---
                - File "<string>", line 2
                  reference to "B"
                The element is skipped because no metatype is given.
                (see https://www.test.org/missing-metatype)
            ''', dict(
                Model='<Model name="Model" uri="http://www.ex.org"/>'))

    def test_missing_metatype_twice(self):
        self.assert_messages_and_models('''
                A = MODEL(name='Model', uri='http://www.ex.org')
                A.B
                A.B
            ''', '''
                #1 --- ERROR: missing-metatype ---
                - File "<string>", line 2
                  reference to "B"
                - File "<string>", line 3
                  reference to "B"
                The element is skipped because no metatype is given.
                (see https://www.test.org/missing-metatype)
            ''', dict(
                Model='<Model name="Model" uri="http://www.ex.org"/>'))

    def test_multiple_metatypes(self):
        self.assert_messages_and_models('''
                A = MODEL(name='Model', uri='http://www.ex.org')
                A.B = CONCRETE_CLASS()
                A.B
                A.B = ABSTRACT_CLASS()
            ''', '''
                #1 --- ERROR: multiple-metatypes ---
                - File "<string>", line 2
                  CONCRETE_CLASS "B"
                - File "<string>", line 4
                  ABSTRACT_CLASS "B"
                The element is skipped because multiple metatypes are given: AbstractClass, ConcreteClass.
                (see https://www.test.org/multiple-metatypes)
            ''', dict(
                Model='<Model name="Model" uri="http://www.ex.org"/>'))

    # TODO
    # def test_invalid_member_type(self):
    #     self.assert_messages_and_models('''
    #             A = MODEL(name='A', uri='http://www.ex.org/A')
    #             A.B = MODEL(name='B', uri='http://www.ex.org/B')
    #         ''', '''
    #             #1 --- ERROR: invalid-member-type ---
    #             - File "<string>", line 1
    #             - File "<string>", line 2
    #             Model A.B is skipped because Model A cannot have a member of this type.
    #             see file:///E:/s/w/p/specificator/doc/.build/html/dsl.html#invalid-member-type
    #         ''')

    def test_invalid_member_type(self):
        self.assert_messages_and_models('''
                A = MODEL(name='A', uri='http://www.ex.org')
                A.B = DATA_PROPERTY(lower=0, upper=1, type=BUILTIN.String)
            ''', '''
                #1 --- ERROR: invalid-member-type ---
                - File "<string>", line 2
                  DATA_PROPERTY "B"
                The element is skipped because its owner of type Model cannot have a member of type DataProperty.
                (see https://www.test.org/invalid-member-type)
            ''', dict(
                A='<Model name="A" uri="http://www.ex.org"/>'))

    def test_invalid_member_type_propagation(self):
        self.assert_messages_and_models('''
                A = MODEL(name='A', uri='http://www.ex.org')
                A.B = DATA_PROPERTY()
                A.B.C = DATA_PROPERTY()
            ''', '''
                #1 --- ERROR: invalid-member-type ---
                - File "<string>", line 2
                  DATA_PROPERTY "B"
                - File "<string>", line 3
                  reference to "B"
                The element is skipped because its owner of type Model cannot have a member of type DataProperty.
                (see https://www.test.org/invalid-member-type)
                
                #2 --- ERROR: owner-skipped ---
                - File "<string>", line 3
                  DATA_PROPERTY "C"
                causes: #1
                The element is skipped because its owner is skipped.
                (see https://www.test.org/owner-skipped)
            ''', dict(
                A='<Model name="A" uri="http://www.ex.org"/>'))

    def test_multiple_attribute_values(self):
        self.assert_messages_and_models('''
                A = MODEL(name='A', uri='http://www.ex.org')
                A.B = CONCRETE_CLASS()
                A.C = CONCRETE_CLASS(superTypes=[A.B])
                A.C = CONCRETE_CLASS(superTypes=[])
            ''', '''
                #1 --- ERROR: multiple-parameter-values ---
                - File "<string>", line 3
                  CONCRETE_CLASS "C"
                - File "<string>", line 4
                  CONCRETE_CLASS "C"
                The value for parameter "superTypes" is given several times. The parameter is ignored.
                (see https://www.test.org/multiple-parameter-values)
            ''', dict(A='''
                <Model name="A" uri="http://www.ex.org">
                  <ConcreteClass name="B"/>
                  <ConcreteClass name="C"/>
                </Model>'''))

    def test_cyclic_dependency_types(self):
        self.assert_messages_and_models('''
                A = MODEL(name='A', uri='http://www.ex.org')
                A.B = CONCRETE_CLASS(superTypes=[A.C])
                A.C = CONCRETE_CLASS(superTypes=[A.B])
            ''', '''
                #1 --- ERROR: cyclic-dependency ---
                - A.B
                  - File "<string>", line 2
                    CONCRETE_CLASS "B"
                  - File "<string>", line 3
                    reference to "B"
                - A.C
                  - File "<string>", line 2
                    reference to "C"
                  - File "<string>", line 3
                    CONCRETE_CLASS "C"
                Due to a cyclic dependency, these elements cannot be built.
                (see https://www.test.org/cyclic-dependency)
            ''', dict(
                A='<Model name="A" uri="http://www.ex.org"/>'))

    def test_templates_type_error(self):
        self.assert_messages_and_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = ABSTRACT_CLASS(templates=1)
            ''', '''
                #1 --- ERROR: invalid-parameter-type ---
                - File "<string>", line 2
                  ABSTRACT_CLASS "A"
                Parameter "templates": got 1 of type int, but value type must be list.
                (see https://www.test.org/invalid-parameter-type)
            ''', dict(
                Model='''
                    <Model name="Model" uri="http://www.ex.org">
                      <AbstractClass name="A"/>
                    </Model>'''))


    def test_templates_item_type_error(self):
        self.assert_messages_and_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = ABSTRACT_CLASS(templates=[1, 2])
            ''', '''
                #1 --- ERROR: invalid-parameter-type ---
                - File "<string>", line 2
                  ABSTRACT_CLASS "A"
                Parameter "templates[0]": got 1 of type int, but value type must be Property.
                (see https://www.test.org/invalid-parameter-type)
                
                #2 --- ERROR: invalid-parameter-type ---
                - File "<string>", line 2
                  ABSTRACT_CLASS "A"
                Parameter "templates[1]": got 2 of type int, but value type must be Property.
                (see https://www.test.org/invalid-parameter-type)
            ''', dict(
                Model='''
                    <Model name="Model" uri="http://www.ex.org">
                      <AbstractClass name="A"/>
                    </Model>'''))

    def test_skipped_element_used_templates(self):
        self.assert_messages_and_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = ABSTRACT_CLASS(templates=[Model.B])
            ''', '''
                #1 --- ERROR: missing-metatype ---
                - File "<string>", line 2
                  reference to "B"
                The element is skipped because no metatype is given.
                (see https://www.test.org/missing-metatype)
                
                #2 --- ERROR: skipped-element-used ---
                - File "<string>", line 2
                  ABSTRACT_CLASS "A"
                causes: #1
                Parameter "templates[0]": 'Model.B' is no valid value: It has been skipped.
                (see https://www.test.org/skipped-element-used)
            ''', dict(
                Model='''
                    <Model name="Model" uri="http://www.ex.org">
                      <AbstractClass name="A"/>
                    </Model>'''))


    def test_skipped_element_used_templates_2(self):
        self.assert_messages_and_models('''
                PROPERTY_TEMPLATES = TEMPLATE_MODEL(name='PropertyTemplates')
                PROPERTY_TEMPLATES.A = ABSTRACT_CLASS()
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = ABSTRACT_CLASS(templates=[PROPERTY_TEMPLATES.A])
            ''', '''
                #1 --- ERROR: invalid-member-type ---
                - File "<string>", line 2
                  ABSTRACT_CLASS "A"
                - File "<string>", line 4
                  reference to "A"
                The element is skipped because its owner of type TemplateModel cannot have a member of type AbstractClass.
                (see https://www.test.org/invalid-member-type)
                
                #2 --- ERROR: skipped-element-used ---
                - File "<string>", line 4
                  ABSTRACT_CLASS "A"
                causes: #1
                Parameter "templates[0]": 'PropertyTemplates.A' is no valid value: It has been skipped.
                (see https://www.test.org/skipped-element-used)
            ''', dict(
                Model='''
                    <Model name="Model" uri="http://www.ex.org">
                      <AbstractClass name="A"/>
                    </Model>'''))


    def test_empty(self):
        self.assert_models('', {})


    def test_abstract_class_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = ABSTRACT_CLASS()
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <AbstractClass name="A"/>
                </Model>'''))


    def test_abstract_data_type_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = ABSTRACT_DATA_TYPE()
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <AbstractDataType name="A"/>
                </Model>'''))


    def test_aggregated_data_type_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = AGGREGATED_DATA_TYPE()
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <AggregatedDataType name="A"/>
                </Model>'''))


    def test_class_parameter_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.B = CONCRETE_CLASS()
                Model.A = CONCRETE_CLASS()
                Model.A.C = CLASS_PARAMETER(type=Model.B)
                Model.A.P = COMPOSITION_PROPERTY(lower=0, upper=3, type=Model.A.C)
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <ConcreteClass name="A">
                    <ClassParameter name="C">
                      <ClassReference type="/B"/>
                    </ClassParameter>
                    <CompositionProperty name="P" isOrdered="false" lower="0" upper="3">
                      <ClassReference type="/A.C"/>
                    </CompositionProperty>
                  </ConcreteClass>
                  <ConcreteClass name="B"/>
                </Model>'''))


    def test_composition_property_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.example.org/Model')
                Model.Wheel = CONCRETE_CLASS()
                Model.Car = CONCRETE_CLASS()
                Model.Car.Wheels = COMPOSITION_PROPERTY(lower=0, upper=4, type=Model.Wheel)
            ''', dict(Model='''
                <Model name="Model" uri="http://www.example.org/Model">
                  <ConcreteClass name="Car">
                    <CompositionProperty name="Wheels" isOrdered="false" lower="0" upper="4">
                      <ClassReference type="/Wheel"/>
                    </CompositionProperty>
                  </ConcreteClass>
                  <ConcreteClass name="Wheel"/>
                </Model>'''))


    def test_concrete_class_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = CONCRETE_CLASS()
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <ConcreteClass name="A"/>
                </Model>'''))


    def test_data_property_for_aggregated_data_type_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.example.org/Model')
                Model.WheelSpecification = AGGREGATED_DATA_TYPE()
                Model.WheelSpecification.Text = DATA_PROPERTY(lower=0, upper=1, type=BUILTIN.String)
            ''', dict(Model='''
                <Model name="Model" uri="http://www.example.org/Model">
                  <AggregatedDataType name="WheelSpecification">
                    <DataProperty name="Text" isOrdered="false" isUnique="false" lower="0" upper="1">
                      <DataTypeReference type="Builtin/String"/>
                    </DataProperty>
                  </AggregatedDataType>
                </Model>'''))


    def test_data_property_for_class_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.example.org/Model')
                Model.Car = CONCRETE_CLASS()
                Model.Car.Color = DATA_PROPERTY(lower=0, upper=1, type=BUILTIN.String)
            ''', dict(Model='''
                <Model name="Model" uri="http://www.example.org/Model">
                  <ConcreteClass name="Car">
                    <DataProperty name="Color" isOrdered="false" isUnique="false" lower="0" upper="1">
                      <DataTypeReference type="Builtin/String"/>
                    </DataProperty>
                  </ConcreteClass>
                </Model>'''))


    def test_data_type_parameter_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = AGGREGATED_DATA_TYPE()
                Model.A.P = DATA_PROPERTY(lower=0, upper=1, type=Model.A.C)
                Model.B = ABSTRACT_DATA_TYPE()
                Model.A.C = DATA_TYPE_PARAMETER(type=Model.B)
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <AggregatedDataType name="A">
                    <DataTypeParameter name="C">
                      <DataTypeReference type="/B"/>
                    </DataTypeParameter>
                    <DataProperty name="P" isOrdered="false" isUnique="false" lower="0" upper="1">
                      <DataTypeReference type="/A.C"/>
                    </DataProperty>
                  </AggregatedDataType>
                  <AbstractDataType name="B"/>
                </Model>'''))


    def test_enumeration_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = ENUMERATION()
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <Enumeration name="A"/>
                </Model>'''))


    def test_enumeration_literal_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.A = ENUMERATION()
                Model.A.B = ENUMERATION_LITERAL()
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <Enumeration name="A">
                    <EnumerationLiteral name="B"/>
                  </Enumeration>
                </Model>'''))


    def test_model_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
            ''', dict(
                Model='<Model name="Model" uri="http://www.ex.org"/>'))


    def test_package_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.ex.org')
                Model.Package = PACKAGE()
            ''', dict(Model='''
                <Model name="Model" uri="http://www.ex.org">
                  <Package name="Package"/>
                </Model>'''))


    def test_reference_property_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.example.org/Model')
                Model.CarPort = CONCRETE_CLASS()
                Model.Car = CONCRETE_CLASS()
                Model.Car.LocatedIn = REFERENCE_PROPERTY(lower=0, upper=1, oppositeLower=0, oppositeUpper=1, type=Model.CarPort)
            ''', dict(Model='''
                <Model name="Model" uri="http://www.example.org/Model">
                  <ConcreteClass name="Car">
                    <ReferenceProperty name="LocatedIn" isOrdered="false" isUnique="false" lower="0" oppositeLower="0" oppositeUpper="1" upper="1">
                      <ClassReference type="/CarPort"/>
                    </ReferenceProperty>
                  </ConcreteClass>
                  <ConcreteClass name="CarPort"/>
                </Model>'''))


    def test_template_model_default(self):
        self.assert_models('''
                Model = MODEL(name='Model', uri='http://www.example.org/Model')
                TemplateModel = TEMPLATE_MODEL(name='TemplateModel', uri='http://www.example.org/TemplateModel')
                TemplateModel.P = DATA_PROPERTY(lower=0, upper=1, type=BUILTIN.String)
                Model.A = CONCRETE_CLASS(templates=[TemplateModel.P])
            ''', dict(Model='''
                <Model name="Model" uri="http://www.example.org/Model">
                  <ConcreteClass name="A">
                    <DataProperty name="P" isOrdered="false" isUnique="false" lower="0" upper="1">
                      <DataTypeReference type="Builtin/String"/>
                    </DataProperty>
                  </ConcreteClass>
                </Model>'''))

    
    
    




from tomlkit import table
from dexpi.specificator.formatter import Formatter

def get_default_config():
    config = table()
    config.add('diagrams', True)
    return config

def make(*args, **kwargs):
    from .make import make
    make(*args, **kwargs)

FORMATTER = Formatter(
    'html', get_default_config, make)



import argparse
import datetime
import os
import pathlib
import sys

from lxml import etree

import tomlkit

import dexpi.specificator

from .dsl_reader import DslReader
from .formatter import get_formatter_by_name, Formatter

from pnb.mcl.metamodel import standard as metamodel

_FORMATTER_BY_NAME = get_formatter_by_name()


def make(model_src, documentation_src, formatters: list[Formatter], cache_dir) -> None:
    
    from pnb.mcl.metamodel import standard as mm

    out_dir = pathlib.Path(os.getcwd()) / '.build'
    out_dir = documentation_src.parent / '.build'
    dsl_reader = DslReader(model_src, cache_dir)
    model_set = dsl_reader.model_set

    
    
    
    if 0:
        
        path = pathlib.Path(r'E:\s\w\p\skas\src\python\dexpilib\dexpilib\test\test_io\test_proteus\test_reader\data\ref\other\mini_pid.built.1.4.xml')
      #  path = pathlib.Path(r'E:\s\w\p\skas\src\python\dexpilib\dexpilib\test\test_io\test_proteus\test_reader\data\ref\other\C01V04-VER.EX02.proteus.xml')
        path = pathlib.Path(r'E:\s\w\p\skas\src\python\dexpilib\dexpilib\test\test_io\test_proteus\test_reader\data\ref\CURRENT\C01.xml')
        
        converter = Converter(
            model_by_name,
            path)
        
        
        converted = converter.result
        from pnb.mcl.io.xml import XmlExporter
        
        print(converted)
        
        
        model = metamodel.Model(
            name='DexpiReferenceExampleC01',
            uri='http://www.example.org/DexpiReferenceExampleC01')
        model.add(converted)
        
        
        
        xml = XmlExporter(model).xml
        
        path.with_suffix('.dexpi20.xml').write_bytes(etree.tostring(xml, pretty_print=True, encoding='utf-8'))
        
        
        
        
        
        
        
        
       # print('OK')
        import sys
        sys.exit()

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    

    
    if False:
        # TODO: remove
        from dexpi.specificator.utils.uml_diagrams import make_diagram
       # from specificator.uml.uml_diagrams import make_uml_diagram as make_diagram
        texes = [r'\documentclass{article}',
                 r'\usepackage{tikz}\usetikzlibrary{arrows,arrows.meta}',
                 r'\begin{document}']

        for element in [
              #  model_by_name['Core'].Note,
              #  model_by_name['Core'].QualifiedValue,
                model_set.search_unique('ProcessStep'),
                model_set.search_unique('Distilling'),
                model_set.search_unique('ConceptualModel', type=metamodel.Class),
              #  model_by_name['Core'].PhysicalQuantities.LengthUnit
           ]:
            
            
            out = out_dir / 'latex' / (element.name + '.tex')
            tex = make_diagram([element], 'tex', out=out, usages_by_class=None)
            texes.append(tex)
            
            
            
        texes.append(r'\end{document}')
        
        
        
        
        pathlib.Path(r'E:\s\test.tex').write_text('\n\n'.join(texes), encoding='utf-8')
            
            
        import subprocess
        subprocess.check_call(
            ['pdflatex',
             r'E:\s\test.tex'],
            cwd=r'E:\s')
                
        return
        
        
        
        for model in model_by_name.values():
        
            for item in model:
                if isinstance(item, metamodel.Type):
                    out = OUT_DIR / 'uml' / (item.name + '.svg')
                    uml_diagrams.make_uml_diagram(
                        [item], out)
                for item in item:
                    if isinstance(item, metamodel.Type):
                        out = OUT_DIR / 'uml' / (item.name + '.svg')
                        uml_diagrams.make_uml_diagram(
                            [item], out)

        return


    if 0:
        # TODO add config support
        try:
            with pathlib.Path(CONFIG_PATH).open('r', encoding='utf-8') as file_in:
                config = tomlkit.load(file_in)
        except Exception as error:
            TODO
    else:
        config = {}
        
  #  return

    for formatter in formatters:
        
        print(f'\nFormatter: {formatter.name}')
        print(out_dir.absolute())
        formatter_config = config.get(formatter, {})
        formatter_config['sphinx_project'] = str(documentation_src)
        formatter_config['index_by_path'] = dsl_reader.index_by_path
        if cache_dir is not None:
            formatter_config['cache_dir'] = str(cache_dir)
        formatter_config['out_dir'] = str(out_dir)
        
        
        

        try:
            formatter.make(model_set, formatter_config)
            print('OK')
        except Exception as error:
            raise
            print()
            print(f'*** {formatter.name} failed ***')
            print(error)
            print()


def get_argument_parser(specification_src=None, documentation_src=None):
    parser = argparse.ArgumentParser(prog='PROG')
    subparsers = parser.add_subparsers()

    make_parser = subparsers.add_parser('make', help='Make output.')
    make_parser.set_defaults(command='make')

    if not specification_src:
        make_parser.add_argument(
            'specification_src',
            metavar='spec',
            help='''
                The directory that contains the SpeciPy sources of the information model.
                Alternatively, the path of a single SpeciPy file.''')

    if not documentation_src:
        make_parser.add_argument(
            '--documentation_src',
            metavar='doc',
            help='The directory that contains a Sphinx project; required for HTML and PDF output.')

    make_parser.add_argument(
        'formatters',
        metavar='formats',
        nargs='*',
        default=None,
        choices=sorted(get_formatter_by_name()),
        help='''
            One or more formats for the output. If omitted, output for all supported formats
            is created.''')

    config_parser = subparsers.add_parser('config', help='create_config help')
    config_parser.set_defaults(command='config')
    config_parser.add_argument('out', help='file name of the config file')

    return parser


def config(out):
    out = pathlib.Path(out).absolute()

    if out.is_file():
        TODO
    elif out.exists():
        TODO

    specificator_config = tomlkit.document()
    specificator_config.add(tomlkit.comment(
        'This is an auto-generated default configuration for the DEXPI Specificator. '
        'Adapt to your needs.'))

    for formatter in _FORMATTER_BY_NAME.values():
        specificator_config.add(formatter.name, (formatter.get_default_config()))

    out.parent.mkdir(parents=True, exist_ok=True)
    # TODO: check encoding
    with out.open('w', encoding='utf-8') as file_out:
        tomlkit.dump(specificator_config, file_out)

    print(f'Default configuration written to {out}.')


def command_line(model_src=None, documentation_src=None, cache_dir=None, args=None):

    print(f'dexpi.specificator {dexpi.specificator.__version__}')
    print()

    if args is None:
        args = sys.argv[1:]

    if not args:
        print('**********************************\n'
              'Assume default command (make all).\n'
              'Run with -h to see more options.  \n'
              '**********************************')
        args = ['make']

    parser = get_argument_parser(model_src, documentation_src)
    arguments = vars(parser.parse_args(args))
    command = arguments.pop('command')

    match command:
        case 'make':

            if model_src:
                arguments['model_src'] = model_src
            if documentation_src:
                arguments['documentation_src'] = documentation_src
            if arguments['formatters']:
                arguments['formatters'] = [
                    _FORMATTER_BY_NAME[name] for name in arguments['formatters']]
            else:
                arguments['formatters'] = list(_FORMATTER_BY_NAME.values())
            arguments['cache_dir'] = cache_dir
            make(**arguments)
        case 'config':
            config(**arguments)
        case _:
            assert False, f'unknown command: {command}'




from __future__ import annotations 

import importlib
import pathlib
import typing as T

if T.TYPE_CHECKING:
    from pnb.mcl.metamodel import standard as metamodel

class Formatter:
    def __init__(self, name, get_default_config, make):
        self.name = name
        self._get_default_config = get_default_config
        self._make = make
        
    def get_default_config(self):
        return self._get_default_config()
    
    def make(self, model_set: metamodel.ModelSet, config: dict):
        actual_config = self.get_default_config()
        actual_config.update(config or {})
        actual_config['out_dir'] = str(pathlib.Path(actual_config['out_dir']) / self.name)
        self._make(model_set, actual_config)

    
_FORMATTER_BY_NAME = None
    
# TODO: rename
def get_formatter_by_name():
    global _FORMATTER_BY_NAME
    if _FORMATTER_BY_NAME is None:
        temp_retval = {}
        from . import formatters
        formatters_dir = pathlib.Path(formatters.__path__[0])
        for dir_entry in formatters_dir.iterdir():
            if dir_entry.is_dir():
                module_name = dir_entry.name
            elif dir_entry.is_file() and dir_entry.suffixes == ['.py']:
                module_name = dir_entry.name[:-3]
            else:
                continue
            module = importlib.import_module(f'.{module_name}', formatters.__name__)
            formatter = getattr(module, 'FORMATTER', None)
            if formatter:
                assert isinstance(formatter, Formatter)
                assert formatter.name not in temp_retval
                temp_retval[formatter.name] = formatter
        _FORMATTER_BY_NAME = temp_retval
    return _FORMATTER_BY_NAME
